from invoke import task
from .docker_tasks import docker_exec


@task
def django(c, port=8000):
    docker_exec(c, "./manage.py runserver 0:{}".format(port))


@task
def gunicorn(c, port=8000):
    gunicorn_config = c.config.get("gunicorn_config", "/usr/local/app/gunicorn.py")
    if gunicorn_config:
        config = f"--config {gunicorn_config}"
    else:
        config = ""
    wsgi_app = c.config.get("wsgi_app", None)
    if wsgi_app is None:
        app = c.config.app
        wsgi_app = f"{app}.wsgi:application"
    docker_exec(c, f"/usr/local/bin/gunicorn {config} -b :{port} {wsgi_app}")


@task
def celery(c):
    app = c.config.app
    celery_args = c.config.get("celery_args", "-P solo -c1")
    docker_exec(c, f"celery worker --app={app} {celery_args}")


@task
def djshell(c):
    docker_exec(c, "./manage.py shell")


@task
def test(c, test=""):
    docker_exec(c, "./manage.py test {}".format(test))


@task
def migrate(c):
    docker_exec(c, "./manage.py migrate")


@task
def manage(c, command):
    docker_exec(c, "./manage.py {}".format(command))


@task
def makemessages(c, language=None):
    if language:
        languages = [language]
    else:
        languages = c.config.translations.languages
    extra_params = c.config.translations.get("extra_params", [])
    for lang in languages:
        docker_exec(c, "./manage.py makemessages -l {} {}".format(lang, " ".join(extra_params)))


@task
def compilemessages(c, language=None):
    docker_exec(c, "./manage.py compilemessages" + ("-l {}".format(language) if language else ""))


@task
def create_su(c, username="admin", email="testing@gogames.co"):
    docker_exec(c, "./manage.py createsuperuser --username {} --email {}".format(username, email))


@task
def coverage(c):
    docker_exec(c, "coverage run --source=. manage.py test")
    docker_exec(c, "coverage html")



import os
import re
import requests
from invoke import task


def _get_aws_token(c):
    token = os.getenv("AWS_TOKEN")
    if not token:
        token = c.run("aws ecr get-authorization-token --output text "
                      "--query 'authorizationData[].authorizationToken'", hide=True).stdout.strip()
    return token


def _get_gcloud_token(c):
    token = os.getenv("GCLOUD_TOKEN")
    if not token:
        token = c.run("gcloud auth print-access-token", hide=True).stdout.strip()
    return token


def _version_to_int(version):
    """Converts a version number into an integer number, so it can be sorted

    >>> _version_to_int("0.1.1")
    1001
    >>> _version_to_int("1.2.3")
    1002003
    >>> _version_to_int("2001")
    2001
    >>> _version_to_int("latest")
    0
    """
    if version == "latest":
        return 0
    components = version.split(".")
    ret = 0
    for i, comp in enumerate(components):
        ret += int(comp) * (1000 ** (len(components) - (i + 1)))
    return ret


def _registry_type(registry):
    if "amazonaws" in registry:
        return "aws"
    elif "gcr.io" in registry:
        return "googlecloud"
    elif "icr.io" in registry:
        return "ibmcloud"
    elif registry == "":
        return "dockerhub"
    else:
        return "unknown"


def _join(registry, image):
    if not registry:
        return image
    return "{}/{}".format(registry, image)


def _auth_headers(c, registry):
    if _registry_type(registry) == "aws":
        token = _get_aws_token(c)
        return dict(headers={'Authorization': 'Basic {}'.format(token)})
    elif _registry_type(registry) == "googlecloud":
        token = _get_gcloud_token(c)
        return dict(auth=("oauth2accesstoken", token))
    else:
        return {}


def _get_last_version_from_local_docker(c, registry, image):
    registry_image = _join(registry, image)
    output = c.run(f"docker image ls {registry_image}", hide="out")
    # Black magic explanation: skips first line (header), 2nd field is version
    tags = [re.split(" +", l)[1] for l in output.stdout.splitlines()[1:]]
    return sorted(tags, key=_version_to_int)[-1]


def _get_last_version(c, registry, image):
    if _registry_type(registry) in ("ibmcloud", "dockerhub"):
        # fallback, don't know how to get tabs from ibmcloud registry
        return _get_last_version_from_local_docker(c, registry, image)

    url = 'https://{}/v2/{}/tags/list'.format(registry, image)
    r = requests.get(url, **_auth_headers(c, registry))
    r.raise_for_status()
    tags = r.json()['tags']
    if len(tags) == 100:
        raise RuntimeError(
            "Error, the response has 100 tags, we hit the limit and paging not supported, "
            "you should remove some tags in ECR console"
        )
    return sorted(tags, key=_version_to_int)[-1]


def _get_next_version(c, registry, image):
    registry, image = _default_registry_image(c, registry, image)
    version = _get_last_version(c, registry, image)
    parts = version.split('.')
    parts[-1] = str(int(parts[-1]) + 1)
    return '.'.join(parts)


def _default_registry_image(c, registry, image):
    if not registry:
        registry = c.config.registry

    if not image:
        image = c.config.image

    return registry, image


@task
def last_version(c, registry=None, image=None):
    registry, image = _default_registry_image(c, registry, image)
    print(_get_last_version(c, registry, image))


@task
def next_version(c, registry=None, image=None):
    registry, image = _default_registry_image(c, registry, image)
    print(_get_next_version(c, registry, image))


def docker_exec(c, command, container=None, pty=True, envs={}):
    container = container or c.config.container
    run_command = "docker exec "
    if pty:
        run_command += "-it "
    for env_var, env_value in envs.items():
        run_command += f"--env {env_var}={env_value} "

    for k, env_value in os.environ.items():
        if k.startswith("DOCKEREXEC_"):
            env_var = k.split('_', 1)[1]
            run_command += f"--env {env_var}={env_value} "

    c.run("{} {} {}".format(run_command, container, command), pty=pty)


@task
def docker_put(c, source, target, container=None):
    container = container or c.config.container
    c.run(f"docker cp {source} {container}:{target}")


@task
def docker_get(c, source, target, container=None):
    container = container or c.config.container
    c.run(f"docker cp {container}:{source} {target}")


@task
def start_dev(c, compose_files="docker-compose.override.dev.yml,docker-compose.override.local-dev.yml"):
    extra_param = ""
    for compose_file in compose_files.split(","):
        if os.path.exists(compose_file):
            extra_param += f"-f {compose_file} "

    c.run(f"docker-compose -f docker-compose.yml {extra_param} up --build -d")


@task
def start(c):
    c.run("docker-compose -f docker-compose.yml --build -d")


@task
def stop(c):
    c.run("docker-compose down")


@task
def shell(c):
    shell = c.config.get("container_shell", "sh")
    docker_exec(c, shell)


@task
def pyshell(c):
    docker_exec(c, "python")


@task
def build(c, registry=None, image=None, version=None):
    registry, image = _default_registry_image(c, registry, image)
    registry_image = _join(registry, image)
    version = version or _get_next_version(c, registry, image)
    c.run("docker build -t {}:{} .".format(registry_image, version))


@task
def push_image(c, registry=None, image=None, version=None):
    registry, image = _default_registry_image(c, registry, image)
    if not version:
        if _registry_type(registry) in ("ibmcloud", "dockerhub"):
            version = _get_last_version_from_local_docker(c, registry, image)
        else:
            version = _get_next_version(c, registry, image)
    if _registry_type(registry) == "aws":
        docker_login_cmd = c.run("aws ecr get-login --no-include-email", hide=True).stdout
        c.run(docker_login_cmd)
    registry_image = _join(registry, image)
    c.run("docker push {}:{}".format(registry_image, version))



import os
import re
import sys
import tempfile
import base64
import yaml
from invoke import task, Failure


def kubectl(c, command, **kargs):
    env = getattr(c.config, "env", {})
    return c.run(f"kubectl {command}", env=env, **kargs)


def get_annotation(c, resource, name, annotation):
    command = f"get {resource} {name} -o=jsonpath='{{.metadata.annotations.{annotation}}}'"
    ret = kubectl(c, command, hide=True)
    return ret.stdout


def _normalize(dirname):
    if not dirname.endswith("/"):
        return dirname + "/"
    return dirname


def _applydelete(c, manifest, action="apply"):
    if manifest == "-":
        manifests = [f.strip() for f in sys.stdin.readlines()]
    else:
        manifests = [f.strip() for f in re.split(r",| |\n", manifest)
                     if f.strip()]
    ret = None
    for mfile in manifests:
        is_file = os.path.isfile(mfile)
        if not is_file and action == "apply":
            print(f"{mfile} does not exists!", file=sys.stderr)
        elif not is_file:  # action == delete
            # Assume it's pod to delete - To support multiple delete from stdin
            ret = kubectl(c, f"{action} pod {mfile}")
        else:
            ret = kubectl(c, f"{action} -f {mfile}")
    return ret


@task
def apply(c, manifest):
    return _applydelete(c, manifest, "apply")


@task
def kdelete(c, manifest, resource=None):
    if manifest == "-" or os.path.isfile(manifest):
        return _applydelete(c, manifest, "delete")
    else:
        resource = resource or "pod"
        kubectl(c, f"delete {resource} {manifest}")


@task
def kdescribe(c, resource, name, namespace=None):
    if namespace:
        namespace = f" -n {namespace}"
    else:
        namespace = ""
    kubectl(c, f"describe {resource} {name}{namespace}")


@task
def krollout(c, name, action="restart", namespace=None):
    if namespace:
        namespace = f" -n {namespace}"
    else:
        namespace = ""
    if "/" not in name:
        name = "deployment/" + name
    kubectl(c, f"rollout {action} {name}{namespace}")


@task
def kc(c, command):
    return kubectl(c, command)


YTT_CREATE_CONFIGMAP = """#@ load("@ytt:data", "data")
apiVersion: v1
kind: ConfigMap
metadata:
  name: #@ data.values.name
  annotations: #@ data.values.annotations
data: #@ data.values.files
"""


YTT_CREATE_SECRET = """#@ load("@ytt:data", "data")
apiVersion: v1
kind: Secret
type: Opaque
metadata:
  name: #@ data.values.name
  annotations: #@ data.values.annotations
data: #@ data.values.files
"""


@task
def config_from_dir(c, name, directory=None, secret=False):
    config = "secret" if secret else "configmap"

    if "/" in name and os.path.isdir(name) and not directory:
        # name parameter is the directory, find the name of the configmap/secret
        name = _normalize(name)
        names = kubectl(c, f"get {config} -o=name", hide=True).stdout.splitlines()
        for n in names:
            n = n.split("/")[1]  # removes resource prefix
            if name == get_annotation(c, config, n, "config-from-dir"):
                directory, name = name, n
                break
        if not directory:
            raise Failure(f"No existing {config} found with config-from-dir={name}")

    if directory is None:
        directory = get_annotation(c, config, name, "config-from-dir")
        if not directory:
            raise Failure(f"Missing directory parameter and annotation not found")

    directory = _normalize(directory)

    template_file = tempfile.NamedTemporaryFile(suffix=".yaml", mode="wt")
    template_file.write(YTT_CREATE_SECRET if secret else YTT_CREATE_CONFIGMAP)
    template_file.flush()

    values = {"name": name, "annotations": {"config-from-dir": directory}, "files": {}}

    for filename in os.listdir(directory):
        file_str = open(os.path.join(directory, filename), "rb" if secret else "rt").read()
        if secret:
            values["files"][filename] = base64.b64encode(file_str)
        else:
            values["files"][filename] = file_str

    return run_ytt(c, template_file.name, values, apply=True)


def _fuzzy_find_pod(c, podname):
    pods = kubectl(c, "get pods", hide="out")
    podnames = [re.split(" +", l)[0] for l in pods.stdout.splitlines()[1:]]
    if podname not in podnames:
        aux = [p for p in podnames if podname in p]
        if aux and len(aux) == 1:
            podname = aux[0]
        elif len(aux) > 1:
            podnames = ", ".join(aux)
            raise Failure(f"More than one pod matching {podname}: {podnames}!")
        else:
            podnames = ", ".join(podnames)
            raise Failure(f"{podname} not found in pods: {podnames}!")
    return podname


@task
def logs(c, podname, zfuzzy=False, app=False, name=False, follow=False, tail=None,
         container=None, max_log_requests=None):
    if zfuzzy:
        podname = _fuzzy_find_pod(c, podname)
    elif app:
        podname = f"-l app={podname}"
    elif name:
        podname = f"-l name={podname}"

    if max_log_requests:
        max_log_requests = f"--max-log-requests {max_log_requests}"
    else:
        max_log_requests = ""

    follow = "--follow" if follow else ""
    tail = f"--tail {tail}" if tail else ""
    container = f"-c {container}" if container else ""
    return kubectl(c, f"logs {podname} {follow} {tail} {container} {max_log_requests}")


@task
def kshell(c, podname, zfuzzy=False, shell="sh"):
    if zfuzzy:
        podname = _fuzzy_find_pod(c, podname)
    return kubectl(c, f"exec -it {podname} -- {shell}", pty=True)


@task
def ktop(c, resource="nodes", cpu=None, memory=None):
    command =f"top {resource}"
    if cpu is None and memory is None:
        return kubectl(c, command)

    cpu = cpu and int(cpu)
    memory = memory and int(memory)

    out = kubectl(c, command, hide=True)
    lines = out.stdout.splitlines()

    for line in lines:
        line_cpu = re.findall(r"\s(\d+)m", line)
        line_memory = re.findall(r"\s(\d+)Mi", line)
        if cpu and line_cpu:
            line_cpu = int(line_cpu[0])
            if cpu > 0 and line_cpu < cpu:
                # Skip lines with LESS than CPU parameter
                continue
            if cpu < 0 and line_cpu > -cpu:
                # Skip lines with MORE than CPU parameter
                continue
        if memory and line_memory:
            line_memory = int(line_memory[0])
            if memory > 0 and line_memory < memory:
                # Skip lines with LESS than memory parameter
                continue
            if memory < 0 and line_memory > -memory:
                # Skip lines with MORE than memory parameter
                continue
        print(line.rstrip("\n"))


@task
def kget(c, resource="pods", grep=None, status=None, keep_header=True, namespace=None,
         name=None, app=None, llist=False):
    if grep:
        hide = "out"
    else:
        hide = None

    if llist:
        keep_header = False

    if namespace is None:
        namespace = ""
    elif namespace == "all":
        namespace = "--all-namespaces"
    else:
        namespace = f"-n={namespace}"

    label_filter = ""
    if app:
        label_filter += f" -l app={app}"
    if name:
        label_filter += f" -l name={name}"

    out = kubectl(c, f"get {resource} {namespace}{label_filter}", hide=hide)
    if hide is None:
        return

    lines = out.stdout.splitlines()

    if grep:
        lines = [l for i, l in enumerate(lines) if grep in l or keep_header and i == 0]

    if llist:
        for l in lines:
            # Only first column, the name
            print(l.split()[0])
    else:
        for l in lines:
            print(l.rstrip("\n"))


def run_ytt(c, template, values=None, output_file=None, apply=False, **kargs):
    f_param = [f"-f {template}"]

    if values is not None:
        values_file = tempfile.NamedTemporaryFile(mode="wt", suffix=".yml")
        values_file.write("#@data/values\n---\n")
        yaml.safe_dump(values, values_file)
        f_param.insert(1, f"-f {values_file.name}")
        values_file.flush()

    if output_file:
        output = f"> {output_file}"
    else:
        output = ""

    f_param = " ".join(f_param)

    if apply and not output_file:
        env = getattr(c.config, "env", {})
        return c.run(f"ytt {f_param} | kubectl apply -f -", env=env, **kargs)

    ret = c.run(f"ytt {f_param} {output}", **kargs)

    if apply:
        ret = kubectl(c, f"apply -f {output_file}", **kargs)
    return ret


@task(iterable=["value_list"])
def ytt(c, template, values=[], output_file=None, apply=False):
    """Processes a manifest with string.Template using os.environ, config.env, and
       config.template_vars[output_file]
    """
    values_dict = dict(v.split("=") for v in values)
    return run_ytt(c, template, values=values_dict, apply=apply)


@task
def generate_templates(c, template_file=None, output_file=None, apply=False):
    templates = c.config.templates

    for template_filename, params in templates.items():
        if template_file and template_filename != template_file:
            continue  # Only process specific template
        default_values = params.get("values", {})
        for out_file_config in params.get("files", []):
            out_file = out_file_config["name"]
            if output_file and output_file != out_file:
                continue  # Only generate specific output
            values = dict(default_values)
            values.update(out_file_config.get("values", {}))
            run_ytt(c, template_filename, values=values,
                    output_file=out_file, apply=apply)
